unit xsdedit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  xmls, xsdins, strs, Menus, StdCtrls, ExtCtrls, db;

type
  TXSDEditor = class(TForm)
    PopupMenu1: TPopupMenu;
    AddShort: TMenuItem;
    AddFull: TMenuItem;
    N1: TMenuItem;
    DelItem: TMenuItem;
    Panel1: TPanel;
    Button1: TButton;
    Button2: TButton;
    XP: TXMLInspector;
    procedure PopupMenu1Popup(Sender: TObject);
    procedure AA1Click(Sender: TObject);
    procedure DelItemClick(Sender: TObject);
  private
   Schema, Node: TXMLNode;
  public
   constructor CreateWithCaption(AOwner: TComponent; const ACaption: string='');
   function  Run(ASchema, ANode: TXMLNode; DS: TDataSet=nil): boolean;
  end;


implementation

{$R *.DFM}

function TXSDEditor.Run(ASchema, ANode: TXMLNode; DS: TDataSet=nil): boolean;
var i: Integer;
begin
 Schema:=ASchema;
 Node:=ANode;
 if Assigned(DS) then
  for i:=0 to DS.FieldCount-1 do
   if Schema.NodebyAttr('name','_'+LowerCase(DS.Fields[i].FieldName))<>nil
    then Node.Attr['_'+LowerCase(DS.Fields[i].FieldName)]:=DS.Fields[i].AsString;
 XP.Schema:=ASchema;
 XP.CurNode:=ANode;
 XP.NodeClicked;
 ShowModal;
 Result:=ModalResult=mrOk;
 if Result and Assigned(DS) and (DS.State in [dsEdit, dsInsert]) then
  for i:=0 to DS.FieldCount-1 do
   if Schema.NodebyAttr('name','_'+LowerCase(DS.Fields[i].FieldName))<>nil
    then DS.Fields[i].AsString:=Node.Attr['_'+LowerCase(DS.Fields[i].FieldName)];
end;

function Title(XN: TXMLNode): string;
begin
 if XN.Attr['ref']<>'' then XN:=XN.Root.NodebyAttr('name',XN.Attr['ref']);
 Result:=XN.Node['xsd:annotation/xsd:documentation'];
 if Result='' then Result:=XN.Attr['name'];
end;

procedure InsertNode(SchemaNode, Node, NewNode: TXMLNode);
var k:integer;

procedure ProcessNode(SchemaNode, Node: TXMLNode; var Start:integer);
var i, k,{minoccurs,} maxoccurs: integer; SN: TXMLNode; ElementName: string;
begin
 {---typed elements---}
 if SchemaNode.Attr['type']<>'' then SchemaNode:=SchemaNode.Root.NodebyAttr('name',SchemaNode.Attr['type']) else
 {---type by ref---}
 if SchemaNode.Attr['ref']<>'' then SchemaNode:=SchemaNode.Root.NodebyAttr('name',SchemaNode.Attr['ref']);
 if SchemaNode=nil then exit;
 for i:=0 to SchemaNode.Count-1 do begin
  if NewNode=nil then exit;
  SN:=SchemaNode[i];
  {---attribute---}
  if SN.Name='xsd:attribute' then else
  {---element---}
  if SN.Name='xsd:element' then begin
   //minOccurs:=0;
   maxOccurs:=1;
   //if SN.Attr['minOccurs']<>'' then minOccurs:=strtoint(SN.Attr['minOccurs']);
   if SN.Attr['maxOccurs']<>'' then begin
    if SN.Attr['maxOccurs']='unbounded' then maxOccurs:=10000
     else maxOccurs:=strtoint(SN.Attr['maxOccurs']);
   end;
   k:=0;
   ElementName:=SN.Attr['name'];
   if ElementName='' then ElementName:=SN.Attr['ref'];
   while (Start<Node.Count) and (Node[Start].Name=ElementName) do begin
    inc(Start);
    inc(k);
   end;
   if Assigned(SN.Parent) and (SN.Parent.Name='xsd:choice') then begin
    if (k<maxOccurs) and ((k>0) or (Node.Count=0)) and (NewNode.Name=ElementName)
      then begin Node.Nodes.Insert(Start, NewNode); NewNode.Parent:=Node; NewNode:=nil; exit end;
   end else begin
    if (k<maxOccurs) and (NewNode.Name=ElementName) then begin
     Node.Nodes.Insert(Start, NewNode);
     NewNode.Parent:=Node;
     NewNode:=nil;
     exit
    end;
   end;
  end else
  {---complexType---}
  if SN.Name='xsd:complexType' then begin
   if SN.Attr['name']<>'' then begin
    Node.Nodes.Insert(Start,NewNode); NewNode.Parent:=Node; NewNode:=nil; exit
   end else ProcessNode(SN, Node, Start);
  end else
  {---simpleContent extension restriction sequence choice---}
  if StrIn(SN.Name, ['xsd:simpleContent', 'xsd:extension', 'xsd:restriction', 'xsd:sequence', 'xsd:choice'])
   then ProcessNode(SN, Node, Start)
 end;
end;

begin
 k:=0;
 ProcessNode(SchemaNode, Node, k)
end;


procedure GetChildList(SchemaNode, Node: TXMLNode; L: TStringList);
var k: integer; ElementName: string;

procedure ProcessNode(SchemaNode, Node: TXMLNode; var Start: integer);
var i, k, minoccurs, maxoccurs: integer; XN, SN: TXMLNode;
begin
 {---typed elements---}
 if SchemaNode.Attr['type']<>'' then SchemaNode:=SchemaNode.Root.NodebyAttr('name', SchemaNode.Attr['type']) else
 {---type by ref---}
 if SchemaNode.Attr['ref']<>'' then SchemaNode:=SchemaNode.Root.NodebyAttr('name', SchemaNode.Attr['ref']);
 if SchemaNode=nil then exit;
 for i:=0 to SchemaNode.Count-1 do begin
  SN:=SchemaNode[i];
  {---attribute---}
  if SN.Name='xsd:attribute' then else
  {---element---}
  if SN.Name='xsd:element' then begin
   if SN.Attr['minOccurs']='' then minOccurs:=1 else
    minOccurs:=strtoint(SN.Attr['minOccurs']);
   if SN.Attr['maxOccurs']<>'' then begin
    if SN.Attr['maxOccurs']='unbounded' then maxOccurs:=10000
     else maxOccurs:=strtoint(SN.Attr['maxOccurs']);
   end else maxOccurs:=1;
   ElementName:=SN.Attr['name'];
   if ElementName='' then ElementName:=SN.Attr['ref'];
   k:=0;
   {---add required elements---}
   if (Start<Node.Count) and (minOccurs>0) and (Node[Start].Name<>ElementName) then begin
     XN:=CreateXMLTemplate(SN, false);
     XN.Name:=ElementName;
     InsertNode(SchemaNode, Node, XN);
   end;
   while (Start<Node.Count) and (Node[Start].Name=ElementName) do begin
    inc(Start);
    inc(k);
   end;
   if Assigned(SN.Parent) and (SN.Parent.Name='xsd:choice') then begin
    if (k<maxOccurs) and ((k>0) or (Node.Count=0))
      then L.AddObject(Title(SN), SN) else L.AddObject('*'+Title(SN), SN);
   end else begin
    if k<maxOccurs then L.AddObject(Title(SN), SN) else L.AddObject('*'+Title(SN), SN);
   end;
  end else
  {---complexType---}
  if SN.Name='xsd:complexType' then begin
   if SN.Attr['name']<>'' then L.AddObject(Title(SN), SN)
    else ProcessNode(SN, Node, Start);
  end else
  {---simpleContent extension restriction sequence choice---}
  if StrIn(SN.Name, ['xsd:simpleContent', 'xsd:extension', 'xsd:restriction', 'xsd:sequence', 'xsd:choice'])
   then ProcessNode(SN, Node, Start)
 end;
end;

begin
 k:=0;
 ProcessNode(SchemaNode, Node, k)
end;

procedure TXSDEditor.PopupMenu1Popup(Sender: TObject);
var L: TStringList; MI: TMenuItem; i:integer; SN: TXMLNode;
begin
 while AddShort.Count>0 do AddShort.Delete(0);
 while AddFull.Count>0 do AddFull.Delete(0);
 L:=TStringList.Create;
 GetChildList(TXSDSchemaEditor(XP.Editor).Schema, TXSDSchemaEditor(XP.Editor).Node, L);
 for i:=0 to L.COunt-1 do begin
  MI:=TMenuItem.Create(self);
  MI.Tag:=integer(L.Objects[i]);
  if StartsWith(L[i],'*') then begin
   MI.Caption:=copy(L[i],2,maxint);
   MI.Enabled:=false;
  end else MI.Caption:=L[i];
  MI.OnClick:=AA1Click;
  AddShort.Add(MI);
  MI:=TMenuItem.Create(self);
  MI.Tag:=integer(L.Objects[i]);
  if StartsWith(L[i],'*') then begin
   MI.Caption:=copy(L[i],2,maxint);
   MI.Enabled:=false;
  end else MI.Caption:=L[i];
  MI.OnClick:=AA1Click;
  AddFull.Add(MI);
 end;
 SN:=TXSDSchemaEditor(XP.Editor).Schema;
 if (SN.Attr['maxOccurs']='1') or (SN.Attr['maxOccurs']='') then DelItem.Enabled:=false
  else DelItem.Enabled:=true;
 L.Free;
end;

procedure TXSDEditor.AA1Click(Sender: TObject);
var SN,XN,Curr: TXMLNode; ElementName:string;
begin
 SN:=TXMLNode((Sender as TMenuItem).Tag);
 if SN.Attr['name']<>'' then ElementName:=SN.Attr['name'] else ElementName:=SN.Attr['ref'];
 Curr:=TXSDSchemaEditor(XP.Editor).Node;
 if (SN.Attr['type']<>'') and (not StartsWith(SN.Attr['type'], 'xsd:'))
   then SN:=SN.Root.NodebyAttr('name', SN.Attr['type']) else
  {---type by ref---}
    if SN.Attr['ref']<>'' then SN:=SN.Root.NodebyAttr('name', SN.Attr['ref']);
 if SN=nil then exit;
 XN:=CreateXMLTemplate(SN, TMenuItem(Sender).Parent=AddFull);
 XN.Name:=ElementName;
 InsertNode(TXSDSchemaEditor(XP.Editor).Schema,Curr, XN);
 XP.NodeClicked;
 XP.NodeClicked;
end;

procedure TXSDEditor.DelItemClick(Sender: TObject);
var SN,Cur: TXMLNode;
begin
 Cur:=XP.CurNode;
 SN:=TXSDSchemaEditor(XP.Editor).Node;
 if Assigned(SN.Parent) and (SN.NodeType=ntNode) then begin
  SN.Parent.Nodes.DeleteNode(SN);
  XP.CurNode:=nil;
  XP.CurNode:=Cur;
 end;
end;


constructor TXSDEditor.CreateWithCaption(AOwner: TComponent; const ACaption: string);
begin
 inherited Create(AOwner);
 if ACaption<>'' then Caption:=ACaption;
end;

end.
